# -*- coding: utf-8 -*-
"""RL_Vanila_Policy_Gradient_Pytorch_Claude_prnbada_v2.ipynb

Automatically generated by Colab.

Original file is located at
    https://colab.research.google.com/drive/1o8zh-M0k4JObc5lrDAICPjeGt-YSBGbJ

# REINFORCE (a.k.a (Vanila) Policy Gradient)

tasks adapted from https://github.com/yandexdataschool/Practical_RL/ week6
"""

!pip install gymnasium

#import gym
import gymnasium as gym
import numpy as np
import pandas as pd

env = gym.make("CartPole-v1", render_mode="rgb_array")

env.reset()

input_dim = env.observation_space.shape[0]
output_dim = env.action_space.n

"""# Building the network for REINFORCE

For REINFORCE algorithm, we'll need a model that predicts action probabilities given states.
"""

import torch
import torch.nn as nn
import torch.optim as optim
from torch.distributions import Categorical
import torch.nn.functional as F
from torch.autograd import Variable

class PolicyNetwork(nn.Module):
    def __init__(self, input_dim, output_dim):
        super(PolicyNetwork, self).__init__()
        self.fc = nn.Sequential(
            nn.Linear(input_dim, 128),
            nn.ReLU(),
            nn.Linear(128, output_dim)
        )

    def forward(self, x):
        return F.softmax(self.fc(x), dim=-1)

# Initialize the policy and optimizer
policy = PolicyNetwork(input_dim, output_dim)
optimizer = optim.Adam(policy.parameters(), lr=0.01)

def select_action(state):
    state = torch.from_numpy(state).float().unsqueeze(0)
    probs = policy(state)
    m = Categorical(probs)
    action = m.sample()
    return action.item(), m.log_prob(action)

def update_policy(rewards, log_probs, log=False):
    discounted_rewards = []
    R = 0
    for r in rewards[::-1]:
        R = r + 0.99 * R
        discounted_rewards.insert(0, R)
    discounted_rewards = torch.tensor(discounted_rewards)
    discounted_rewards = (discounted_rewards - discounted_rewards.mean()) / (discounted_rewards.std() + 1e-9)

    policy_loss = []
    for log_prob, R in zip(log_probs, discounted_rewards):
        policy_loss.append(-log_prob * R)
    policy_loss = torch.cat(policy_loss).sum()

    optimizer.zero_grad()
    policy_loss.backward()
    optimizer.step()
    #policy_loss.backward()
    #optimizer.step()
    #optimizer.zero_grad()

# Training loop
num_episodes = 400
max_steps = 500

for episode in range(num_episodes):
    state, _ = env.reset()
    rewards = []
    log_probs = []

    for step in range(max_steps):
        action, log_prob = select_action(state)
        next_state, reward, terminated, truncated, _ = env.step(action)

        done = terminated or truncated

        rewards.append(reward)
        log_probs.append(log_prob)

        state = next_state

        if done:
            break

    log=False

    update_policy(rewards, log_probs, log)

    if episode % 10 == 0:
        total_reward = sum(rewards)
        print(f'Episode {episode}, Total Reward: {total_reward}')

env.close()
