# -*- coding: utf-8 -*-
"""word2vec_20240511.ipynb

Automatically generated by Colab.

Original file is located at
    https://colab.research.google.com/drive/1q6lgujscGXOyhl6_MRsumBZ6gLlxnOJ9
"""

import urllib.request
urllib.request.urlretrieve("https://itjaeneungnanum.org/data/bbc-text.csv", filename="bbc-text.csv")

import nltk
nltk.download('stopwords')

from nltk.corpus import stopwords
STOPWORDS = set(stopwords.words('english'))
print(STOPWORDS)

# Commented out IPython magic to ensure Python compatibility.
import re, string
import pandas as pd
from time import time
from collections import defaultdict
import spacy
from sklearn.manifold import TSNE
from gensim.models import Word2Vec
# %matplotlib inline

df = pd.read_csv('bbc-text.csv')

df.head()

df.text

def clean_text(text):
    '''Make text lowercase, remove text in square brackets, remove punctuation and remove words containing numbers.'''
    text = text.lower()
    text = re.sub(r'\[.*?\]', '', text)
    text = re.sub(r'[%s]' % re.escape(string.punctuation), '', text)
    text = re.sub(r'\w*\d\w*', '', text)
    # Remove a sentence if it is only one word long
    if len(text) > 2:
        return ' '.join(word for word in text.split() if word not in STOPWORDS)

df_clean = pd.DataFrame(df.text.apply(lambda x: clean_text(x)))

df_clean.head()

#nlp = spacy.load('en', disable=['ner', 'parser']) # disabling Named Entity Recognition for speed
nlp = spacy.load('en_core_web_sm', disable=['ner', 'parser']) # disabling Named Entity Recognition for speed

def lemmatizer(text):
    sent = []
    doc = nlp(text)
    for word in doc:
        sent.append(word.lemma_)
    return " ".join(sent)

df_clean["text_lemmatize"] =  df_clean.apply(lambda x: lemmatizer(x['text']), axis=1)

df_clean.head()

# spaCy uses -PRON- as lemma for personal pronouns,
df_clean['text_lemmatize_clean'] = df_clean['text_lemmatize'].str.replace('-PRON-', '')

sentences = [row.split() for row in df_clean['text_lemmatize_clean']]

word_freq = defaultdict(int)
for sent in sentences:
    for i in sent:
        word_freq[i] += 1
len(word_freq)
print(word_freq)

sorted(word_freq, key=word_freq.get, reverse=True)[:10]

# min_count: minimum number of occurrences of a word in the corpus to be included in the model.
# window: the maximum distance between the current and predicted word within a sentence.
# size: the dimensionality of the feature vectors
# workers: I know my system is having 4 cores,
w2v_model = Word2Vec(sentences,
                     min_count=200,
                     window=5,
                     #size=100,
                     vector_size=100,
                     workers=4)

# Explore the model
w2v_model.wv.most_similar(positive=['economy'])

w2v_model.wv.most_similar(positive=['president'])

# min_count: minimum number of occurrences of a word in the corpus to be included in the model.
# window: the maximum distance between the current and predicted word within a sentence.
# size: the dimensionality of the feature vectors
# workers: I know my system is having 4 cores,
w2v_model = Word2Vec(min_count=200,
                     window=5,
                     #size=100,
                     vector_size=100,
                     workers=4)

# this line of code to prepare the model vocabulary
w2v_model.build_vocab(sentences)

# train word vectors
#w2v_model.train(sentences, total_examples=w2v_model.corpus_count, epochs=w2v_model.iter)
w2v_model.train(sentences, total_examples=w2v_model.corpus_count, epochs=3000)

# As we do not plan to train the model any further,
# we are calling init_sims(), which will make the model much more memory-efficient
w2v_model.init_sims(replace=True)

# Explore the model
w2v_model.wv.most_similar(positive=['economy'])

# how similar are these two words to each other
w2v_model.wv.similarity('company', 'business')

import numpy as np
import matplotlib.pyplot as plt

def tsne_plot(model):
    "Create TSNE model and plot it"
    labels = []
    tokens = []

    #for word in model.wv.vocab:
    for word in list(model.wv.key_to_index.keys()):
        tokens.append(model.wv[word])
        labels.append(word)

    tsne_model = TSNE(perplexity=40, n_components=2, init='pca', n_iter=2500, random_state=23)
    #new_values = tsne_model.fit_transform(tokens)
    new_values = tsne_model.fit_transform(np.array(tokens))

    x = []
    y = []
    for value in new_values:
        x.append(value[0])
        y.append(value[1])

    plt.figure(figsize=(18, 18))
    for i in range(len(x)):
        plt.scatter(x[i],y[i])
        plt.annotate(labels[i],
                     xy=(x[i], y[i]),
                     xytext=(5, 2),
                     textcoords='offset points',
                     ha='right',
                     va='bottom')
    plt.show()

# Use t-SNE to represent high-dimensional data
# and the underlying relationships between vectors in a lower-dimensional space.
tsne_plot(w2v_model)